unit CHColorPicker;

{ ####################################################################################
  TCHColorPicker

  Version   :   1.0.0
  Delphi    :   5 & 6
  Autor     :   Christian Hmmerle
  eMail     :   chaemmerle@uni.de
  Internet  :   http://www.Blue-Xplosion.de (German/English)

  Lizenz    :   Freeware

  Abgeleitet:   TComponent

  History:
  1.0.0 - 21.07.2002    - First Release

  #################################################################################### }

interface

uses
  Windows, Forms, Messages, SysUtils, Classes, Controls, ExtCtrls, StdCtrls, Graphics, ShellApi,
  Math,

  _CHTypes;

type
  TCHColorPicker = class;

  TPickTimer = class(TPersistent)
  private
    FOwner : TCHColorPicker;
    FTimer : TTimer;
  public
    constructor Create(AOwner: TCHColorPicker); virtual;
    destructor Destroy; override;
  end;

  //***********************************************************************************************************

  TCHColorPicker = class(TComponent)
  private
    FOnTimer : TNotifyEvent;
    FPickTimer : TPickTimer;

    FEnabled : Boolean;
    FInterval : Cardinal;

    procedure GetColor(Sender: TObject);
    function DesktopColor(const x, y: integer): TColor;
    procedure ConvertColor(const Color : TColor);
    procedure SetEnabled(const Value: Boolean);
    procedure SetInterval(const Value: Cardinal);
  public
    FColor : TColor;
    FColorString : string;
    FColorHex : string;
    FColorHTML : string;
    FColorRGB_R : Byte;
    FColorRGB_G : Byte;
    FColorRGB_B : Byte;

    FColorHSV_H : Double;
    FColorHSV_S : Double;
    FColorHSV_V : Double;

    FColorHLS_H : Double;
    FColorHLS_L : Double;
    FColorHLS_S : Double;

    FColorCMY_C : Double;
    FColorCMY_M : Double;
    FColorCMY_Y : Double;

    constructor Create(AOwner: TComponent); override;

  published
    property OnTimer : TNotifyEvent read FOnTimer write FOnTimer;
    property Enabled : Boolean read FEnabled Write SetEnabled;
    property Interval : Cardinal read FInterval Write SetInterval;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('CH Pack', [TCHColorPicker]);
end;


{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHColorPicker.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FEnabled := False;
  FInterval := 100;

  FPickTimer := TPickTimer.Create(Self);

end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TPickTimer.Create(AOwner: TCHColorPicker);
begin
  inherited Create;
  FOwner := AOwner;

  FTimer := TTimer.Create(AOwner);
  FTimer.Interval := FOwner.FInterval;
  FTimer.Enabled := FOwner.FEnabled;
  FTimer.OnTimer := FOwner.GetColor;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
destructor TPickTimer.Destroy;
begin
  FTimer.Free;
  inherited Destroy;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHColorPicker.GetColor(Sender: TObject);
var
  CurPos : TPoint;
begin 
  GetCursorPos(CurPos);
  FColor := DesktopColor(CurPos.x, CurPos.y);
  ConvertColor(FColor);

  if Assigned(FOnTimer) then
    FOnTimer(Self);
end;

function TCHColorPicker.DesktopColor(const x, y: integer): TColor;
var
  hDesktop: THandle;
  DC: HDC;
begin
  hDesktop := GetDesktopWindow;
  DC := GetWindowDC(hDesktop);
  try
    Result := GetPixel(DC, x, y);
  finally
    ReleaseDC(hDesktop, DC);
  end;
end;


{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHColorPicker.ConvertColor(const Color: TColor);
var
  tmpColor : TColor;
  Delta, Min, Max :  Double;
  R, G, B : Byte;
begin
  tmpColor := ColorToRGB(Color);

  // HTML
  FColorHTML := Format('#%.2x%.2x%.2x', [(tmpColor) and $FF, (tmpColor shr  8) and $FF, (tmpColor shr 16) and $FF]);

  // RGB
  R := GetRValue(Color);
  G := GetGValue(Color);
  B := GetBValue(Color);
  FColorRGB_R := R;
  FColorRGB_G := G;
  FColorRGB_B := B;

  // Hex
  FColorHex := IntToHex(R, 2) + IntToHex(G, 2) + IntToHex(B, 2);

  // String
  FColorString := ColorToString(Color);

  Min := MinValue( [R, G, B] );
  Max := MaxValue( [R, G, B] );

  //HSV
  try
    FColorHSV_V := Max;

    Delta := FColorHSV_V - Min;

    if FColorHSV_V =  0.0 then
      FColorHSV_S := 0
    else
      FColorHSV_S := Delta / FColorHSV_V;

    if FColorHSV_S  = 0.0 then
      FColorHSV_H := 0
    else
    begin
      if R = FColorHSV_V then
        FColorHSV_H := 60.0 * (G - B) / Delta
      else
      if G = FColorHSV_V then
        FColorHSV_H := 120.0 + 60.0 * (B - R) / Delta
      else
      if  B = FColorHSV_V then
        FColorHSV_H := 240.0 + 60.0 * (R - G) / Delta;

      if   FColorHSV_H < 0.0 then
        FColorHSV_H := FColorHSV_H + 360.0;
    end;
  except
    Abort;
  end;

  // HLS
  try
    FColorHLS_L := (Max + Min) / 2.0;

    if Max = Min then
    begin
      FColorHLS_S := 0.0;
      FColorHLS_H := 0;
    end
    else
    begin
      Delta := Max - Min;

      if FColorHLS_L <= 0.5 then
        FColorHLS_S := Delta / (Max + Min)
      else
        FColorHLS_S := Delta / (2.0 - (Max + Min));

      if R = Max then
        FColorHLS_H := (60.0*(G - B)) / Delta
      else
      if G = Max then
        FColorHLS_H := 120.0 + (60.0*(B - R)) / Delta
      else
      if B = Max then
        FColorHLS_H := 240.0 + (60.0*(R - G)) / Delta;

      if FColorHLS_H < 0 then
        FColorHLS_H := FColorHLS_H + 360.0;
    end;
  except
    Abort;
  end;

  // CMY
  try
    FColorCMY_C := 1.0 - R;
    FColorCMY_M := 1.0 - G;
    FColorCMY_Y := 1.0 - B;
  except
    Abort;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHColorPicker.SetEnabled(const Value: Boolean);
begin
  if FEnabled <> Value then
  begin
    FEnabled := Value;
    FPickTimer.FTimer.Enabled := FEnabled;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHColorPicker.SetInterval(const Value: Cardinal);
begin
  if FInterval <> Value then
  begin
    FInterval := Value;
    if FInterval = 0 then
      FInterval := 1;

    FPickTimer.FTimer.Interval := FInterval;
  end;
end;


end.
