unit CHThreadTimer;

{ ####################################################################################
  TCHThreadTimer

  Version   :   1.0.0
  Delphi    :   5 & 6
  Autor     :   Christian Hmmerle
  eMail     :   chaemmerle@uni.de
  Internet  :   http://www.Blue-Xplosion.de (German/English)

  Lizenz    :   Freeware

  Abgeleitet:   TComponent

  History:
  1.0.0 - 21.07.2002    - First Release

  #################################################################################### }

interface

uses
  Windows, Forms, Messages, SysUtils, Classes, Controls, StdCtrls, Graphics, ShellApi,

  _CHTypes;

type
  TCHThreadTimer = class;

  TExtThread = class(TThread)
  private
    FOwner : TCHThreadTimer;
    FStop : THandle;
  protected
    procedure Execute; override;
  end;

  //***********************************************************************************************************

  TCHThreadTimer = class(TComponent)
  private
    FOnTimer: TNotifyEvent;
    FExtThread : TExtThread;
    FBreakMode : TBreakMode;

    FEnabled : Boolean;
    FAllowZero : Boolean;
    FSynchronized : Boolean;
    FEndThreadWaiting : Boolean;
    FThreadPriority : TThreadPriority;
    FInterval : Cardinal;

    procedure DoTimer;

    procedure SetEnabled(const Value : Boolean);
    procedure SetInterval(const Value : Cardinal);
    procedure SetThreadPriority(const Value : TThreadPriority);
    procedure SetSynchronized(const Value: Boolean);
    procedure SetAllowZero(const Value: Boolean);
    procedure setBreakMode(const Value: TBreakMode);
    procedure SetEndThreadWaiting(const Value: Boolean);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property OnTimer: TNotifyEvent read FOnTimer write FOnTimer;

    property AllowZero: Boolean read FAllowZero Write SetAllowZero;
    property BreakMode : TBreakMode read FBreakMode Write setBreakMode;
    property Enabled: Boolean read FEnabled write SetEnabled;
    property EndThreadWaiting : Boolean read FEndThreadWaiting Write SetEndThreadWaiting;
    property Interval: Cardinal read FInterval write SetInterval;
    property Synchronized: Boolean read FSynchronized write SetSynchronized;
    property ThreadPriority: TThreadPriority read FThreadPriority  write SetThreadPriority;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('CH Pack', [TCHThreadTimer]);
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TExtThread.Execute;
begin
  repeat
    if WaitForSingleObject(FStop, FOwner.Interval) = WAIT_TIMEOUT then
    begin
      if FOwner.Synchronized then
        Synchronize(FOwner.DoTimer)
      else
        FOwner.DoTimer;
    end;
  until Terminated;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
constructor TCHThreadTimer.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FAllowZero := False;
  FEnabled := False;
  FInterval := 1000;
  FSynchronized := True;
  FEndThreadWaiting := False;
  FThreadPriority := tpNormal;
  FBreakMode := thSuspend;

  FExtThread := TExtThread.Create(True);
  FExtThread.FOwner := Self;
  FExtThread.Priority := tpNormal;
  FExtThread.FStop := CreateEvent(nil, False, False, nil);

end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
destructor TCHThreadTimer.Destroy;
begin
  Enabled := False;

  FExtThread.Terminate;
  SetEvent(FExtThread.FStop);
  if FExtThread.Suspended then
    FExtThread.Resume;

  FExtThread.WaitFor;
  CloseHandle(FExtThread.FStop);
  FExtThread.Free;

  inherited Destroy;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHThreadTimer.DoTimer;
begin
  if FEnabled and Assigned(FOnTimer) and not (csDestroying in ComponentState) then
  begin
    try
      FOnTimer(Self);
    except
    end;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHThreadTimer.SetAllowZero(const Value: Boolean);
begin
  if FAllowZero <> Value then
    FAllowZero := Value;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHThreadTimer.setBreakMode(const Value: TBreakMode);
begin
  if FBreakMode <> Value then
    FBreakMode := Value;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHThreadTimer.SetEnabled(const Value: Boolean);
begin
  if Value <> FEnabled then
  begin
    FEnabled := Value;
    // Enable = True
    if FEnabled then
    begin
      if (FInterval > 0) or ((FInterval = 0) and FAllowZero) then
      begin
        SetEvent(FExtThread.FStop);

        if FBreakMode = thSuspend then
          FExtThread.Resume
        else
        begin
          if Assigned(FExtThread) then
            FExtThread.Terminate;
          FExtThread := TExtThread.Create(True);
          FExtThread.FOwner := Self;
          FExtThread.Priority := FThreadPriority;
          FExtThread.FStop := CreateEvent(nil, False, False, nil);
          FExtThread.Resume;
        end;
      end;
    end
    // Enable = False
    else
    begin
      // Suspend
      if FBreakMode = thSuspend then
      begin
        FExtThread.Suspend;
      end
      // Terminate
      else
      begin
        if FEndThreadWaiting then
        begin
          FExtThread.FreeOnTerminate := False;
          FExtThread.Terminate;
          FExtThread.WaitFor;
        end
        else
        begin
          FExtThread.Terminate;
        end;
      end;
    end;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHThreadTimer.SetEndThreadWaiting(const Value: Boolean);
begin
  if FEndThreadWaiting <> Value then
    FEndThreadWaiting := Value;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHThreadTimer.SetInterval(const Value: Cardinal);
var
  tmpEnabled : Boolean;
  tmpInterval : Cardinal;
begin
  if Value <> FInterval then
  begin
    tmpEnabled := FEnabled;
    tmpInterval := FInterval;
    Enabled := False;

    if (FInterval = 0) and (not FAllowZero) then
      FInterval := tmpInterval
    else
      FInterval := Value;

    Enabled := tmpEnabled;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHThreadTimer.SetSynchronized(const Value: Boolean);
begin
  if FSynchronized <> Value then
  begin
    FSynchronized := Value;
  end;
end;

{ ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ }
procedure TCHThreadTimer.SetThreadPriority(const Value: TThreadPriority);
begin
  if FThreadPriority <> Value then
  begin
    FExtThread.Priority := Value;
    FThreadPriority := FExtThread.Priority;
  end;
end;


end.
